// This is the main DLL file.

#include "stdafx.h"

#include "G65DDCAPIWrapper.h"

using namespace G65DDCAPIWrapperSpace;
using namespace System;
using namespace System::Runtime::InteropServices;


//
// ApiHandler
//
HMODULE ourModuleHandle = NULL;
INT32 ourDeviceHandle = -1;
G65DDC_CALLBACKS ourCallbacks;

G65DDC_OPEN_DEVICE ourOpenDevice;
G65DDC_CLOSE_DEVICE ourCloseDevice;
G65DDC_GET_DEVICE_LIST ourGetDeviceList;
G65DDC_IS_DEVICE_CONNECTED ourIsDeviceConnected;
G65DDC_SET_POWER ourSetPower;
G65DDC_SET_FREQUENCY ourSetFrequency;
G65DDC_GET_FREQUENCY ourGetFrequency;
G65DDC_GET_RANGE ourGetRange;
G65DDC_SET_AGC ourSetAGC;
G65DDC_SET_AGC_PARAMS ourSetAGCParams;
G65DDC_SET_DEMODULATOR_MODE ourSetDemodulatorMode;
G65DDC_SET_DEMODULATOR_BANDWIDTH ourSetDemodulatorBandwidth;
G65DDC_CREATE_DDC1 ourCreateDDC1;
G65DDC_CREATE_DDC2 ourCreateDDC2;
G65DDC_SET_DDC ourSetDDC;
G65DDC_GET_DDC ourGetDDC;
G65DDC_START_DDC ourStartDDC;
G65DDC_STOP_DDC ourStopDDC;
G65DDC_START_AUDIO ourStartAudio;
G65DDC_STOP_AUDIO ourStopAudio;
G65DDC_SET_VOLUME ourSetVolume;
G65DDC_GET_VOLUME ourGetVolume;
G65DDC_SET_CALLBACKS ourSetCallbacks;

G65DDCAPIWrapper::G65DDCAPIWrapper()
{
	instance = this; // Used in callback from managed C++ to C#
	devicesCount = 0;

	IntPtr aPointer = Marshal::StringToHGlobalUni(APILIBRARY);
	ourModuleHandle = LoadLibrary((LPCWSTR)aPointer.ToPointer());

	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("Failed to load {0}", APILIBRARY);
		status = STATUS_APILIBRARY_NOT_FOUND;
		return;
	}

	// Retrieve addresses of used API functions
    ourOpenDevice = (G65DDC_OPEN_DEVICE)GetProcAddress(ourModuleHandle, "OpenDevice");
    ourCloseDevice = (G65DDC_CLOSE_DEVICE)GetProcAddress(ourModuleHandle, "CloseDevice");
	ourGetDeviceList = (G65DDC_GET_DEVICE_LIST)GetProcAddress(ourModuleHandle, "GetDeviceList");
	ourIsDeviceConnected = (G65DDC_IS_DEVICE_CONNECTED)GetProcAddress(ourModuleHandle, "IsDeviceConnected");
	ourSetPower = (G65DDC_SET_POWER)GetProcAddress(ourModuleHandle, "SetPower");
	ourSetFrequency = (G65DDC_SET_FREQUENCY)GetProcAddress(ourModuleHandle, "SetFrequency");
	ourGetFrequency = (G65DDC_GET_FREQUENCY)GetProcAddress(ourModuleHandle, "GetFrequency");
	ourGetRange = (G65DDC_GET_RANGE)GetProcAddress(ourModuleHandle, "GetRange");
	ourSetAGC = (G65DDC_SET_AGC)GetProcAddress(ourModuleHandle, "SetAGC");
	ourSetAGCParams = (G65DDC_SET_AGC_PARAMS)GetProcAddress(ourModuleHandle, "SetAGCParams");
    ourSetDemodulatorMode = (G65DDC_SET_DEMODULATOR_MODE)GetProcAddress(ourModuleHandle, "SetDemodulatorMode");
    ourSetDemodulatorBandwidth = (G65DDC_SET_DEMODULATOR_BANDWIDTH)GetProcAddress(ourModuleHandle, "SetDemodulatorFilterBandwidth");
	ourCreateDDC1 = (G65DDC_CREATE_DDC1)GetProcAddress(ourModuleHandle, "CreateDDC1");
	ourCreateDDC2 = (G65DDC_CREATE_DDC2)GetProcAddress(ourModuleHandle, "CreateDDC2");
	ourSetDDC = (G65DDC_SET_DDC)GetProcAddress(ourModuleHandle, "SetDDC");
	ourGetDDC = (G65DDC_GET_DDC)GetProcAddress(ourModuleHandle, "GetDDC");
	ourStartDDC = (G65DDC_START_DDC)GetProcAddress(ourModuleHandle, "StartDDC");
	ourStopDDC = (G65DDC_STOP_DDC)GetProcAddress(ourModuleHandle, "StopDDC");
	ourStartAudio = (G65DDC_START_AUDIO)GetProcAddress(ourModuleHandle, "StartAudio");
	ourStopAudio = (G65DDC_STOP_AUDIO)GetProcAddress(ourModuleHandle, "StopAudio");
	ourGetVolume = (G65DDC_GET_VOLUME)GetProcAddress(ourModuleHandle, "GetVolume");
	ourSetVolume = (G65DDC_SET_VOLUME)GetProcAddress(ourModuleHandle, "SetVolume");
	ourSetCallbacks = (G65DDC_SET_CALLBACKS)GetProcAddress(ourModuleHandle, "SetCallbacks");

	if (!ourOpenDevice || !ourCloseDevice || !ourGetDeviceList || !ourIsDeviceConnected || !ourSetPower ||
		!ourSetFrequency || !ourGetFrequency ||
		!ourSetAGC || !ourSetAGCParams || !ourSetDemodulatorMode || !ourSetDemodulatorBandwidth ||
		!ourCreateDDC1 || !ourCreateDDC2 || !ourSetDDC || !ourGetDDC || !ourStartDDC || !ourStopDDC || 
		!ourStartAudio || !ourStopAudio ||
		!ourGetVolume || !ourSetVolume ||
		!ourSetCallbacks)
	{
		Console::WriteLine("Incorrect {0}", APILIBRARY);
		FreeLibrary(ourModuleHandle);
		ourModuleHandle = NULL;
		status = STATUS_APILIBRARY_INCORRECT;
		return;
	}

	Console::WriteLine("{0} loaded", APILIBRARY);

	status = STATUS_APILIBRARY_LOADED;
}

G65DDCAPIWrapper::~G65DDCAPIWrapper()
{
	FreeLibrary(ourModuleHandle);
}

DWORD G65DDCAPIWrapper::GetStatus()
{
	return status;
}

BOOL G65DDCAPIWrapper::IsConnected(BOOL* theIsConnected)
{
	if (ourDeviceHandle < 0)
	{
		return FALSE;
	}
	else
	{
		return ourIsDeviceConnected(ourDeviceHandle, theIsConnected);
	}
}

String^ G65DDCAPIWrapper::GetDevicePath()
{
	return deviceInfo.devicePath;
}

BYTE G65DDCAPIWrapper::GetInterfaceType()
{
	return deviceInfo.interfaceType;
}

String^ G65DDCAPIWrapper::GetSerialNumber()
{
	return deviceInfo.serialNumber;
}

array<FrequencyRange^>^ G65DDCAPIWrapper::GetRanges()
{
	return deviceInfo.ranges;
}

BOOL G65DDCAPIWrapper::Enumerate()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);
		return FALSE;
	}

	// Let's work with one device
	G65DDC_DEVICE_INFO* aDeviceList;
	UINT32 aDevicesCount;
	BOOL aListResult = ourGetDeviceList(&aDeviceList, &aDevicesCount);
	if (!aListResult || aDevicesCount == 0)
	{
		Console::WriteLine("No available device found");
        return FALSE;
	}
	devicesCount = aDevicesCount;

	G65DDC_DEVICE_INFO aDeviceInfo = aDeviceList[0];

	deviceInfo.devicePath = gcnew String(aDeviceInfo.DevicePath);
	deviceInfo.interfaceType = aDeviceInfo.InterfaceType;
	deviceInfo.serialNumber = gcnew String(aDeviceInfo.SerialNumber);
	deviceInfo.hwVersion = aDeviceInfo.HWVersion;

	BYTE aCount = 3;
	deviceInfo.fwVersion = gcnew array<UINT16>(aCount);
	for (BYTE anIx = 0; anIx < aCount; anIx++)
	{
		deviceInfo.fwVersion[anIx] = aDeviceInfo.FWVersion[anIx];
	}

	deviceInfo.eepromVersion = aDeviceInfo.EEPROMVersion;

	aCount = G65DDC_RANGE_COUNT;
	deviceInfo.ranges = gcnew array<FrequencyRange^>(aCount);
	for (DWORD anIx = 0; anIx < aCount; anIx++)
	{
		deviceInfo.ranges[anIx] = gcnew FrequencyRange();
		deviceInfo.ranges[anIx]->minFrequency = aDeviceInfo.Ranges[anIx].MinFrequency;
		deviceInfo.ranges[anIx]->maxFrequency = aDeviceInfo.Ranges[anIx].MaxFrequency;
	}

	deviceInfo.maxDDC1ChannelCount = aDeviceInfo.MaxDDC1ChannelCount;
	deviceInfo.maxDDC2ChannelCount = aDeviceInfo.MaxDDC2ChannelCount;
	deviceInfo.maxDDC1TypeCount = aDeviceInfo.MaxDDC1TypeCount;
	deviceInfo.maxDDC2TypeCount = aDeviceInfo.MaxDDC2TypeCount;
	deviceInfo.maxDDC2ChannelsPerDDC1Channel = aDeviceInfo.MaxDDC2ChannelsPerDDC1Channel;
	deviceInfo.flags = aDeviceInfo.Flags;

	aCount = 6;
	deviceInfo.MACAddress = gcnew array<BYTE>(aCount);
	for (BYTE anIx = 0; anIx < aCount; anIx++)
	{
		deviceInfo.MACAddress[anIx] = aDeviceInfo.MACAddress[anIx];
	}

	Console::WriteLine("Device #{0} found", deviceInfo.serialNumber);

	return TRUE;
}

BOOL G65DDCAPIWrapper::Open()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);
		return FALSE;
	}

	if (devicesCount == 0)
	{
		Console::WriteLine("No available device found");
		return FALSE;
	}

	IntPtr aPointer = Marshal::StringToHGlobalAnsi(deviceInfo.serialNumber);
	const char* aSerialNumber = static_cast<char*>(aPointer.ToPointer());
	// Open device
	ourDeviceHandle = ourOpenDevice(aSerialNumber);
	Marshal::FreeHGlobal(aPointer);

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Unable to open device");
		return FALSE;
	}

	Console::WriteLine("Device opened");

	// Set power
    ourSetPower(ourDeviceHandle, TRUE);

	// Example initialization.
	// Create DDC1
	pin_ptr<UINT32> aDDCChannelIdPtr = &DDC1ChannelId;
	ourCreateDDC1(ourDeviceHandle, aDDCChannelIdPtr);

	// Create DDC2 for DDC1, and prepare map of channels (virtual RXs), that may be useful for C# example
	UINT32	aDDC2ChannelId,
			aCount = 1; // Only one channel for this example
	channels = gcnew array<UINT32>(aCount); // Map of channel indexes to DDC2 channel Ids
	DDC2ChannelIds = gcnew array<UINT32>(deviceInfo.maxDDC1ChannelCount + deviceInfo.maxDDC2ChannelCount); // Map of DDC2 channel Ids to channel Ids
	for (DWORD anIx = 0; anIx < aCount; anIx++)
	{
		ourCreateDDC2(ourDeviceHandle, DDC1ChannelId, &aDDC2ChannelId);
		
		// Mapping
		channels[anIx /* Index correspoding to channel */] = aDDC2ChannelId;
		DDC2ChannelIds[aDDC2ChannelId] = anIx;
	}

	// Set DDC type of DDC1 to 1 (DDC Bandwidth = 24000, SampleRate = 32 kHz, BitsPerSample = 32)
	ourSetDDC(ourDeviceHandle, DDC1ChannelId, 1);

	// Set AGC parameters: attack = 0.02 s, decay = 0.5 s, ref. level = -15 dB
	ourSetAGCParams(ourDeviceHandle, aDDC2ChannelId, 0.020, 0.500, -15.0);

	// Enable AGC
	ourSetAGC(ourDeviceHandle, aDDC2ChannelId, TRUE);

	// Set AM mode
	ourSetDemodulatorMode(ourDeviceHandle, aDDC2ChannelId, G65DDC_MODE_AM);

	// Set demodulator filter bandwidth to 8 kHz
	ourSetDemodulatorBandwidth(ourDeviceHandle, aDDC2ChannelId, 8000);

	return TRUE;
}

BOOL G65DDCAPIWrapper::Close()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	// Unregister all callback functions
    ourSetCallbacks(ourDeviceHandle, NULL, 0);

    ourSetPower(ourDeviceHandle, FALSE);

	BOOL aResult = ourCloseDevice(ourDeviceHandle);

	if (aResult)
	{
		ourDeviceHandle = -1;

		Console::WriteLine("Device closed");
	}

	return aResult;
}

BOOL G65DDCAPIWrapper::SetFrequency(UINT32 theChannel, INT32 theFrequency)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourSetFrequency(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theFrequency);
}

BOOL G65DDCAPIWrapper::GetFrequency(UINT32 theChannel, INT32* theFrequency)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourGetFrequency(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theFrequency);
}

BOOL G65DDCAPIWrapper::GetRange(UINT32* theRange)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourGetRange(ourDeviceHandle, theRange);
}

BOOL G65DDCAPIWrapper::RegisterCallbacks(DDC2PreprocessedStreamCallback^ theDDC2PreprocessedStreamCallback, AudioStreamCallback^ theAudioStreamCallback)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	ddc2PreprocessedStreamCallback = theDDC2PreprocessedStreamCallback; // Callback to C#
	audioStreamCallback = theAudioStreamCallback; // Callback to C#

	// DDC2PreprocessedStream-callback from unmanaged to managed C++
	NotificationDDC2PreprocessedStreamCallbackDelegate^ aDDC2PreprocessedStreamDelegate = gcnew NotificationDDC2PreprocessedStreamCallbackDelegate(NotificationDDC2PreprocessedStreamCallback);
	GCHandle aHandle = GCHandle::Alloc(aDDC2PreprocessedStreamDelegate);
	IntPtr aPointer = Marshal::GetFunctionPointerForDelegate(aDDC2PreprocessedStreamDelegate);
	G65DDC_DDC2_PREPROCESSED_STREAM_CALLBACK aDDC2PreprocessedStreamCallback = static_cast<G65DDC_DDC2_PREPROCESSED_STREAM_CALLBACK>(aPointer.ToPointer());

	// AudioStreamCallback from unmanaged to managed C++
	NotificationAudioStreamCallbackDelegate^ anAudioStreamCallbackDelegate = gcnew NotificationAudioStreamCallbackDelegate(NotificationAudioStreamCallback);
	aHandle = GCHandle::Alloc(anAudioStreamCallbackDelegate);
	aPointer = Marshal::GetFunctionPointerForDelegate(anAudioStreamCallbackDelegate);
	G65DDC_AUDIO_STREAM_CALLBACK anAudioStreamCallback = static_cast<G65DDC_AUDIO_STREAM_CALLBACK>(aPointer.ToPointer());

	// Force garbage collection cycle to prove that the delegate does not get disposed
	GC::Collect();

	// Registration of callbacks
	ourCallbacks.ADCSnapshotCallback = NULL;
    ourCallbacks.DDC1StreamCallback = NULL;
    ourCallbacks.DDC1PlaybackStreamCallback = NULL;
    ourCallbacks.DDC2StreamCallback = NULL;
    ourCallbacks.DDC2PreprocessedStreamCallback = aDDC2PreprocessedStreamCallback;
    ourCallbacks.AudioStreamCallback = anAudioStreamCallback;
    ourCallbacks.AudioPlaybackStreamCallback = NULL;


	// Register callbacks
    ourSetCallbacks(ourDeviceHandle, &ourCallbacks, 0);

	return TRUE;
}

BOOL G65DDCAPIWrapper::GetDDC1(UINT32* theDDCIndex, G65DDCInfo^ theDDCInfo)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	G65DDC_DDC_INFO aDDCInfo;
	if (!ourGetDDC(ourDeviceHandle, DDC1ChannelId, theDDCIndex, &aDDCInfo))
	{
		Console::WriteLine("Getting DDC Info failed");

		return FALSE;
	}

	theDDCInfo->sampleRate = aDDCInfo.SampleRate;
	theDDCInfo->bandwidth = aDDCInfo.Bandwidth;
	theDDCInfo->bitsPerSample = aDDCInfo.BitsPerSample;

	return TRUE;
}

BOOL G65DDCAPIWrapper::StartDDC1(UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartDDC(ourDeviceHandle, DDC1ChannelId, theSamplesPerBuffer);
}

BOOL G65DDCAPIWrapper::StopDDC1()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopDDC(ourDeviceHandle, DDC1ChannelId);
}

BOOL G65DDCAPIWrapper::GetDDC2(UINT32 theChannel, UINT32* theDDCIndex, G65DDCInfo^ theDDCInfo)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	G65DDC_DDC_INFO aDDCInfo;
	if (!ourGetDDC(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theDDCIndex, &aDDCInfo))
	{
		Console::WriteLine("Getting DDC Info failed");

		return FALSE;
	}

	theDDCInfo->sampleRate = aDDCInfo.SampleRate;
	theDDCInfo->bandwidth = aDDCInfo.Bandwidth;
	theDDCInfo->bitsPerSample = aDDCInfo.BitsPerSample;

	return TRUE;
}

BOOL G65DDCAPIWrapper::StartDDC2(UINT32 theChannel, UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartDDC(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theSamplesPerBuffer);
}

BOOL G65DDCAPIWrapper::StopDDC2(UINT32 theChannel)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopDDC(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */);
}

BOOL G65DDCAPIWrapper::StartAudio(UINT32 theChannel, UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartAudio(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theSamplesPerBuffer);
}

BOOL G65DDCAPIWrapper::StopAudio(UINT32 theChannel)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopAudio(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */);
}

BOOL G65DDCAPIWrapper::SetVolume(UINT32 theChannel, BYTE theVolume)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourSetVolume(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theVolume);
}

BOOL G65DDCAPIWrapper::GetVolume(UINT32 theChannel, BYTE* theVolume)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourGetVolume(ourDeviceHandle, channels[theChannel] /* Get DDC2 channel Id */, theVolume);
}

void G65DDCAPIWrapper::NotificationDDC2PreprocessedStreamCallback(UINT32 theDDC2ChannelId, const FLOAT* theBuffer, UINT32 theCount, FLOAT thePeak, FLOAT theRMS, DWORD_PTR theUserData)
{
	// Callback to C#
	instance->ddc2PreprocessedStreamCallback(DDC2ChannelIds[theDDC2ChannelId] /* Get virtual RX channel */, theBuffer, theCount, thePeak, theRMS);
}

void G65DDCAPIWrapper::NotificationAudioStreamCallback(UINT32 theDDC2ChannelId, UINT32 theStage, const FLOAT* theBuffer, UINT32 theCount, DWORD_PTR theUserData)
{
	// In example only buffers of final audio samples are used (audio gain, audio filter and volume are applied)
	if (theStage == G65DDC_AUDIO_STREAM_CALLBACK_STAGE_2)
	{
		// Callback to C#
		instance->audioStreamCallback(DDC2ChannelIds[theDDC2ChannelId] /* Get virtual RX channel */, theStage, theBuffer, theCount);
	}
}