#include "G65DDCAPI.h"

#pragma once

using namespace System;

namespace G65DDCAPIWrapperSpace
{
	//
	// Frequency Range
	//
	public ref class FrequencyRange
    {
		public:
		
			UINT32 minFrequency;
			UINT32 maxFrequency;
    };  


	//
	// Device Info
	//
	public ref class G65DDCAPIWrapperDeviceInfo
	{
		public:

			String^ devicePath;
			BYTE interfaceType;
			String^ serialNumber;
			UINT16 hwVersion;
			array<UINT16>^ fwVersion;
			BYTE eepromVersion;
			array<FrequencyRange^>^ ranges;
			UINT32 maxDDC1ChannelCount;    
			UINT32 maxDDC2ChannelCount;
			UINT32 maxDDC1TypeCount;
			UINT32 maxDDC2TypeCount;   
			UINT32 maxDDC2ChannelsPerDDC1Channel;
			UINT32 flags;
			array<BYTE>^ MACAddress;
	};


	//
	// DDC Info
	//
	public ref class G65DDCInfo
	{
		public:

			UINT32 sampleRate;
			UINT32 bandwidth;
			UINT32 bitsPerSample;
	};


	//
	// API Wrapper
	//
	public ref class G65DDCAPIWrapper
	{
		public:

			static String^ APILIBRARY = gcnew String("G65DDCAPI.dll");

			static const DWORD STATUS_APILIBRARY_LOADED = 100;
			static const DWORD STATUS_APILIBRARY_NOT_FOUND = 200;
			static const DWORD STATUS_APILIBRARY_INCORRECT = 300;
			static const DWORD STATUS_DEVICE_CREATINGFAILED = 400;

			static const BYTE DEVICE_INTERFACE_PCI = G65DDC_INTERFACE_TYPE_PCIE;
			static const BYTE DEVICE_INTERFACE_USB2 = G65DDC_INTERFACE_TYPE_USB2;
			static const BYTE DEVICE_INTERFACE_USB3 = G65DDC_INTERFACE_TYPE_USB3;

			static const BYTE RANGE_1 = G65DDC_RANGE_1;
			static const BYTE RANGE_2 = G65DDC_RANGE_2;

			G65DDCAPIWrapper();
			~G65DDCAPIWrapper();

			delegate void DDC2PreprocessedStreamCallback(UINT32 channel, const FLOAT* buffer, UINT32 count, FLOAT peak, FLOAT rms); // Callback to C#
			delegate void AudioStreamCallback(UINT32 channel, UINT32 stage, const FLOAT* buffer, UINT32 count); // Callback to C#

			static void NotificationDDC2PreprocessedStreamCallback(UINT32 ddc2ChannelId, const FLOAT* buffer, UINT32 count, FLOAT peak, FLOAT rms, DWORD_PTR userData); // Callback for Notification (to unmanaged C++)
			delegate void NotificationDDC2PreprocessedStreamCallbackDelegate(UINT32, const FLOAT*, UINT32, FLOAT, FLOAT, DWORD_PTR); // Callback from Notification (from unmanaged C++)
			static void NotificationAudioStreamCallback(UINT32 ddc2ChannelId, UINT32 stage, const FLOAT* buffer, UINT32 count, DWORD_PTR userData); // Callback for Notification (to unmanaged C++)
			delegate void NotificationAudioStreamCallbackDelegate(UINT32, UINT32, const FLOAT*, UINT32, DWORD_PTR); // Callback from Notification (from unmanaged C++)

			DWORD GetStatus();
			BOOL IsConnected(BOOL* isConnected);
			String^ GetDevicePath();
			BYTE GetInterfaceType();
			String^ GetSerialNumber();
			array<FrequencyRange^>^ GetRanges();

			BOOL Enumerate();
			BOOL Open();
			BOOL Close();
			BOOL SetFrequency(UINT32 channel, INT32 frequency);
			BOOL GetFrequency(UINT32 channel, INT32* frequency);
			BOOL GetRange(UINT32* range);
			BOOL RegisterCallbacks(DDC2PreprocessedStreamCallback^ ddc2PreprocessedStreamCallback, AudioStreamCallback^ audioStreamCallback);
			BOOL GetDDC1(UINT32* ddcIndex, G65DDCInfo^ ddcInfo);
			BOOL StartDDC1(UINT32 samplesPerBuffer);
			BOOL StopDDC1();
			BOOL GetDDC2(UINT32 channel, UINT32* ddcIndex, G65DDCInfo^ ddcInfo);
			BOOL StartDDC2(UINT32 channel, UINT32 samplesPerBuffer);
			BOOL StopDDC2(UINT32 channel);
			BOOL StartAudio(UINT32 channel, UINT32 samplesPerBuffer);
			BOOL StopAudio(UINT32 channel);
			BOOL SetVolume(UINT32 channel, BYTE volume);
			BOOL GetVolume(UINT32 channel, BYTE* volume);

		private:

			static G65DDCAPIWrapper^ instance;
			DWORD status;

			int devicesCount;

			G65DDCAPIWrapperDeviceInfo deviceInfo;

			UINT32 DDC1ChannelId;
			static array<UINT32>^ channels; // Maps index of channels (virtual RXs) to DDC2 channel Ids
			static array<UINT32>^ DDC2ChannelIds; // Maps DDC2 channel Ids to virtual RX indexes

			DDC2PreprocessedStreamCallback^ ddc2PreprocessedStreamCallback;
			AudioStreamCallback^ audioStreamCallback;
	};
}
